----------------------------------------------------------------------------------
-- Sample code to drive Numato Lab's 5 inch TFT LCD Expansion Module
-- 
-- By Cacahuete   (community.numato.com/members/cacahuete.94/)
--    2015/11/22
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;



entity driver_5inch_lcd is
    Port ( clk        : in  STD_LOGIC;
           rst_n      : in  STD_LOGIC;
           lcd_clk    : out STD_LOGIC;
           lcd_den    : out STD_LOGIC;
           lcd_hsync  : out STD_LOGIC;
           lcd_vsync  : out STD_LOGIC;
           lcd_red    : out STD_LOGIC_VECTOR(7 downto 0);
           lcd_green  : out STD_LOGIC_VECTOR(7 downto 0);
           lcd_blue   : out STD_LOGIC_VECTOR(7 downto 0);
           SegEna     : out STD_LOGIC_VECTOR(2 downto 0));
end driver_5inch_lcd;



architecture RTL of driver_5inch_lcd is

  --Screen resolution (in pixels)
  constant  HAA         : integer   := 480;       --Horizontal Active Area
  constant  VAA         : integer   := 272;       --Vertical Active Area

  --Horizontal parameters (in pixels)
  constant  HFP         : integer   :=  40;       --Front Porch
  constant  HBP         : integer   :=  40;       --Back Porch
  constant  HSP         : integer   :=   1;       --Sync Pulse
  constant  HSYNC       : integer   :=   0;       --Value of h_cnt where sync pulse occurs
  constant  HAA_L       : integer   := HFP;       --Left boundary of vertical active area
  constant  HAA_R       : integer   := HFP + HAA; --Right boundary of vertical active Area
  constant  H_CNT_MAX   : integer   := HFP + HAA + HBP + HSP;
  
  --Vertical parameters (in lines)
  constant  VFP         : integer   :=   8;       --Front Porch
  constant  VBP         : integer   :=   8;       --Back Porch
  constant  VSP         : integer   :=   1;       --Sync Pulse
  constant  VSYNC       : integer   :=   0;       --Value of v_cnt where sync pulse occurs
  constant  VAA_L       : integer   := VFP;       --Left boundary of vertical active area
  constant  VAA_R       : integer   := VFP + VAA; --Right boundary of vertical active area
  constant  V_CNT_MAX   : integer   := VFP + VAA + VBP + VSP;

  --Pixel clock: 10MHz
  constant  CLK_CNT_HALF    : integer   :=  5; --half period
  signal    clk_cnt         : integer range 0 to CLK_CNT_HALF-1;
  signal    clk_pixel       : std_logic; --pixel clock to output to display
  signal    clk_pixel_d1    : std_logic; --latched pixel clock for edge detection
  signal    clk_pixel_fall  : std_logic; --pulse when falling edge on clk_pixel
  
  --We are in the display area when disp_ena = '1'
  signal    disp_ena    : std_logic;
  
  --Pixel counters
  signal    h_cnt       : integer range 0 to H_CNT_MAX;
  signal    v_cnt       : integer range 0 to V_CNT_MAX;
  
  --Color counter to generate a test card
  signal    color_cnt   : unsigned(7 downto 0);
  signal    color_h_cnt : unsigned(3 downto 0);
  signal    color_v_cnt : unsigned(3 downto 0);
  
begin

  --Shutdown seven segments...
  SegEna <= (others => '1');
  

  lcd_clk   <= clk_pixel;
  lcd_den   <= disp_ena;
  
  --Pixel clock edge detection
  clk_pixel_fall  <= not clk_pixel and clk_pixel_d1;


  --[ CODE responsible for color generation ]----------
  color_cnt <= color_v_cnt & color_h_cnt;
  
  --Fast and dirty, there might be a cleaner way ;)
  -- All the 1px border is red
  -- Right and bottom are blue
  -- Left and top are green

  lcd_blue    <=  (others => '0') when disp_ena = '0' else
                  (others => '0') when ((h_cnt = HFP+1)     or (h_cnt = HFP+HAA)     or  (v_cnt = VFP) or (v_cnt = VFP+VAA-1)) else
                  (others => '1') when ((h_cnt = HFP+HAA-1) or (v_cnt = VFP+VAA-2)) else
                  (others => '0') when ((h_cnt = HFP+2)     or (v_cnt = VFP+1))     else
                  std_logic_vector(color_cnt(7 downto 6)) & std_logic_vector(color_cnt(1 downto 0)) & X"0";

  lcd_green   <=  (others => '0') when disp_ena = '0' else
                  (others => '0') when ((h_cnt = HFP+1)     or (h_cnt = HFP+HAA)     or  (v_cnt = VFP) or (v_cnt = VFP+VAA-1)) else
                  (others => '0') when ((h_cnt = HFP+HAA-1) or (v_cnt = VFP+VAA-2)) else
                  (others => '1') when ((h_cnt = HFP+2)     or (v_cnt = VFP+1))     else
                  std_logic_vector(color_cnt(3 downto 0)) & X"0";

  lcd_red     <=  (others => '0') when disp_ena = '0' else
                  (others => '1') when ((h_cnt = HFP+1)     or (h_cnt = HFP+HAA)     or  (v_cnt = VFP) or (v_cnt = VFP+VAA-1)) else
                  (others => '0') when ((h_cnt = HFP+HAA-1) or (v_cnt = VFP+VAA-2)) else
                  (others => '0') when ((h_cnt = HFP+2)     or (v_cnt = VFP+1))     else
                  std_logic_vector(color_cnt(7 downto 4)) & X"0";

  color_h_cnt <=  (others => '0') when disp_ena = '0' else
                  X"0" when h_cnt <   30+HFP else
                  X"1" when h_cnt >=  30+HFP and h_cnt < 60 +HFP    else
                  X"2" when h_cnt >=  60+HFP and h_cnt < 90 +HFP    else
                  X"3" when h_cnt >=  90+HFP and h_cnt < 120+HFP    else
                  X"4" when h_cnt >= 120+HFP and h_cnt < 150+HFP    else
                  X"5" when h_cnt >= 150+HFP and h_cnt < 180+HFP    else
                  X"6" when h_cnt >= 180+HFP and h_cnt < 210+HFP    else
                  X"7" when h_cnt >= 210+HFP and h_cnt < 240+HFP    else
                  X"8" when h_cnt >= 240+HFP and h_cnt < 270+HFP    else
                  X"9" when h_cnt >= 270+HFP and h_cnt < 300+HFP    else
                  X"A" when h_cnt >= 300+HFP and h_cnt < 330+HFP    else
                  X"B" when h_cnt >= 330+HFP and h_cnt < 360+HFP    else
                  X"C" when h_cnt >= 360+HFP and h_cnt < 390+HFP    else
                  X"D" when h_cnt >= 390+HFP and h_cnt < 420+HFP    else
                  X"E" when h_cnt >= 420+HFP and h_cnt < 450+HFP    else
                  X"F";
  
  color_v_cnt <=  (others => '0') when disp_ena = '0' else
                  X"0" when v_cnt <   17+VFP else
                  X"1" when v_cnt >=  17+VFP and v_cnt <  34+VFP    else
                  X"2" when v_cnt >=  34+VFP and v_cnt <  51+VFP    else
                  X"3" when v_cnt >=  51+VFP and v_cnt <  68+VFP    else
                  X"4" when v_cnt >=  68+VFP and v_cnt <  85+VFP    else
                  X"5" when v_cnt >=  85+VFP and v_cnt < 102+VFP    else
                  X"6" when v_cnt >= 102+VFP and v_cnt < 119+VFP    else
                  X"7" when v_cnt >= 119+VFP and v_cnt < 136+VFP    else
                  X"8" when v_cnt >= 136+VFP and v_cnt < 153+VFP    else
                  X"9" when v_cnt >= 153+VFP and v_cnt < 170+VFP    else
                  X"A" when v_cnt >= 170+VFP and v_cnt < 187+VFP    else
                  X"B" when v_cnt >= 187+VFP and v_cnt < 204+VFP    else
                  X"C" when v_cnt >= 204+VFP and v_cnt < 221+VFP    else
                  X"D" when v_cnt >= 221+VFP and v_cnt < 238+VFP    else
                  X"E" when v_cnt >= 238+VFP and v_cnt < 255+VFP    else
                  X"F";
  --[/ CODE responsible for color generation ]----------


  --Generate a pixel clock
  clock_generate_proc : process(clk, rst_n)
  begin
    if rst_n = '0' then
      clk_pixel     <= '0';
      clk_pixel_d1  <= '0';
      clk_cnt       <= 0;
      
    elsif rising_edge(clk) then
      --Latch pixel clock to detect edges
      clk_pixel_d1  <= clk_pixel;
      
      --Invert pixel clock each half period
      if clk_cnt = CLK_CNT_HALF-1 then
        clk_pixel   <= not clk_pixel;
        clk_cnt     <= 0;
      else
        clk_cnt     <= clk_cnt+1;
      end if;
    end if;
  end process;


  --Counters management
  pixel_counters_proc : process(clk, rst_n)
  begin
    if rst_n = '0' then
      h_cnt     <= 0;
      v_cnt     <= 0;
      
    elsif rising_edge(clk) then
      --We work on falling edge of pixel clock
      if clk_pixel_fall = '1' then
        --Horizontal counter
        if h_cnt = H_CNT_MAX then
          h_cnt   <= 0;
        else
          h_cnt   <= h_cnt + 1;
        end if;
        
        --Vertical counter
        if h_cnt = H_CNT_MAX then
          if v_cnt = V_CNT_MAX then
            v_cnt <= 0;
          else
            v_cnt <= v_cnt + 1;
          end if;
        end if;
      end if;
    end if;
  end process;
  
  
  --Horizontal and Vertical sync pulses
  sync_pulse_proc : process(clk, rst_n)
  begin
    if rst_n = '0' then
      lcd_hsync <= '1';
      lcd_vsync <= '1';
      disp_ena  <= '0';
      
    elsif rising_edge(clk) then
      --We work on falling edge of pixel clock
      if clk_pixel_fall = '1' then
      
        --Horizontal sync signal
        if  (h_cnt = HSYNC) then
          lcd_hsync <= '0';
        else
          lcd_hsync <= '1';
        end if;
        
        --Vertical sync signal
        if  (v_cnt = VSYNC) then
          lcd_vsync <= '0';
        else
          lcd_vsync <= '1';
        end if;
        
        --Display area
        if    ((h_cnt >= HAA_L and h_cnt < HAA_R) and (v_cnt >= VAA_L and v_cnt < VAA_R)) then
          disp_ena  <= '1';
        else
          disp_ena <= '0';
        end if;

      end if;
    end if;
  end process;

end RTL;
